
import logging
import warnings

try:
    from hyperopt import fmin, hp, tpe, Trials
    from hyperopt.tpe import suggest as tpe_suggest
    from hyperopt.rand import suggest as rand_suggest
    from hyperopt.atpe import suggest as atpe_suggest
    using_ultraopt = True
except ImportError:
    warnings.warn('If you wish to include hyperopt you must pip install hyperopt')
    using_ultraopt = False

if using_ultraopt:
    logging.getLogger('hyperopt').setLevel(logging.ERROR)

    def hyperopt_cube(objective, n_trials, n_dim, with_count=False, algo=None):
        """ Minimize a function on the cube using HyperOpt, and audit # of function calls
           :param objective:    function on (0,1)^n_dim
           :param n_trials:     Guideline for function evaluations
           :param n_dim:
           :param with_count:
           :return:
        """
        logging.getLogger('hyperopt').setLevel(logging.ERROR)

        assert algo is not None, 'provide algo'
        hp_space = dict([('u' + str(i), hp.uniform('u' + str(i), 0, 1)) for i in range(n_dim)])

        global feval_count
        feval_count = 0

        def _objective(hps):
            global feval_count
            feval_count += 1
            us = [hps['u' + str(i)] for i in range(n_dim)]
            return objective(us)

        trls = Trials()
        res = fmin(_objective, space=hp_space, algo=tpe.suggest, trials=trls, max_evals=n_trials, show_progressbar=False)
        best_x = [trls.best_trial['misc']['vals']['u' + str(i)][0] for i in range(n_dim)]
        best_val = trls.best_trial['result']['loss']
        return (best_val, best_x, feval_count) if with_count else (best_val, best_x)


    def hyperopt_atpe_cube(objective, n_trials, n_dim, with_count=False):
        return hyperopt_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, algo=atpe_suggest)


    def hyperopt_tpe_cube(objective, n_trials, n_dim, with_count=False):
        return hyperopt_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, algo=tpe_suggest)


    def hyperopt_rand_cube(objective, n_trials, n_dim, with_count=False):
        return hyperopt_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, algo=rand_suggest)


    HYPEROPT_OPTIMIZERS = [hyperopt_atpe_cube, hyperopt_rand_cube, hyperopt_tpe_cube]
else:
    HYPEROPT_OPTIMIZERS = []


if __name__ == '__main__':
    from humpday.objectives.classic import CLASSIC_OBJECTIVES

    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in HYPEROPT_OPTIMIZERS:
            print(optimizer(objective, n_trials=250, n_dim=6, with_count=True))



import optuna
from optuna.logging import CRITICAL
from humpday.objectives.classic import CLASSIC_OBJECTIVES
import logging

logging.getLogger('optuna').setLevel(logging.ERROR)


def optuna_cube_factory(objective, n_trials, n_dim, with_count=False, method=None):

    if method.lower()=='random':
        sampler = optuna.samplers.RandomSampler()
    elif method.lower()=='cmaes':
        sampler = optuna.samplers.CmaEsSampler()
    elif method.lower()=='tpe':
        sampler = optuna.samplers.TPESampler()
    else:
        raise ValueError('random, cmaes, tpe or grid please')

    global feval_count
    feval_count = 0

    def cube_objective(trial):
        global feval_count
        us = [ trial.suggest_float('u'+str(i),0,1) for i in range(n_dim)]
        feval_count += 1
        return objective(us)

    optuna.logging.set_verbosity(CRITICAL)
    study = optuna.create_study(sampler=sampler)
    study.optimize(cube_objective,n_trials=n_trials)

    best_x = [ study.best_params['u'+str(i)] for i in range(n_dim) ]

    if with_count:
        return study.best_value, best_x, feval_count
    else:
        return study.best_value, best_x


def optuna_random_cube(objective, n_trials,n_dim, with_count=False):
    return optuna_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='random')


def optuna_random_cube_clone(objective, n_trials,n_dim, with_count=False): # duplicate
    return optuna_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='random')


def optuna_random_cube_clone_1(objective, n_trials,n_dim, with_count=False): # duplicate
    return optuna_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='random')


def optuna_random_cube_clone_2(objective, n_trials,n_dim, with_count=False): # duplicate
    return optuna_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='random')


def optuna_cmaes_cube(objective, n_trials, n_dim, with_count=False):
    return optuna_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='cmaes')


def optuna_tpe_cube(objective, n_trials, n_dim, with_count=False):
    return optuna_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='tpe')


OPTUNA_OPTIMIZERS = [ optuna_cmaes_cube, optuna_tpe_cube, optuna_random_cube, optuna_random_cube_clone, optuna_random_cube_clone_1,
                      optuna_random_cube_clone_2 ]


if __name__=='__main__':
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in OPTUNA_OPTIMIZERS:
            print((optimizer(objective, n_trials=250, n_dim=6, with_count=True)))



from humpday.comparison.naming import M  #Silly https://www.linkedin.com/posts/tedpetrou_pythonquiz-activity-6766019617098924033-6jIL

from humpday.optimizers.alloptimizers import OPTIMIZERS
from humpday.objectives.allobjectives import OBJECTIVES
from humpday.objectives.allobjectives import CLASSIC_OBJECTIVES
from humpday.comparison.suggestions import suggest, recommend
from humpday.optimizers.meta import minimize






M = ['January','February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December']


def optimizer_name(solver):
    return solver.__name__.replace('_cube','')


def objective_name(objective):
    return objective.__name__.replace('_cube','')


from deap.benchmarks import h1
from timemachines.common.plotting import mesh2d_
import numpy as np
import math

# Work in progress


def rotate(origin, point, angle):
    """
    Rotate a point counterclockwise by a given angle around a given origin.

    The angle should be given in radians.
    """
    ox, oy = origin
    px, py = point

    qx = ox + math.cos(angle) * (px - ox) - math.sin(angle) * (py - oy)
    qy = oy + math.sin(angle) * (px - ox) + math.cos(angle) * (py - oy)
    return qx, qy


def coords(zs,dim=2):
    """ Unflatten into list of tuples """
    return [ np.array(zs[i:i+2]) for i in range(0,len(zs),dim) ]

def hump(xs,j):
    """ Loosely based on h1 function """
    origin = 0,0
    xs_rotated = rotate(origin=origin,point=xs,angle=j*math.pi/2)
    xs_dilated = [ (x*100) for i,x in enumerate(xs_rotated)]
    return -h1(individual=xs_dilated)[0]+0.0005*np.linalg.norm(xs_dilated)


def dilated_hump(xs,j):
    """ Loosely based on h1 function """
    origin = 0,0
    xs_rotated = rotate(origin=origin,point=xs,angle=j*math.pi/2)
    xs_dilated = [ (x*100)*(i+1) for i,x in enumerate(xs_rotated)]
    return -h1(individual=xs_dilated)[0]+0.0005*np.linalg.norm(xs_dilated)

def hump_selector(xs,zs,ws,offset):
    """ Finds nearest center and returns hump function
       xs position in plane
       zs listing of centers of humps
       ws weights of humps
       offset
    """
    centers = coords( zs, 2)
    ys = [ xs-c+offset for c in centers ]
    distances = [ (np.linalg.norm(ys_),i) for i, ys_ in enumerate(ys) ]
    sort_dist = sorted(distances,reverse=False)
    i_close = sort_dist[0][1]
    y = ys[i_close]
    w = ws[i_close]
    return w*hump(y,j=i_close)


if __name__=='__main__':
    ws = np.array([1.0,2.0,1.3])
    zs = np.array([0.25, 0.25, -0.25, -0.25, 0.15, -0.11])
    offset = (0,0)
    mesh2d_(hump_selector,zs,ws,offset)


from humpday.objectives.classic import CLASSIC_OBJECTIVES
from skopt import gp_minimize
global feval_count
from copy import deepcopy
import traceback

GP_INTERPRETATIONS = dict([(ipg, {'initial_point_generator':ipg}) for ipg in ["random", "sobol", "hammersly", "lhs"]])
GP_INTERPRETATIONS.update({  'lcbexplore':{"acq_func": "LCB", "kappa":3.0},
                             'lcb':{"acq_func":"LCB","kappa":1.96},
                             'lcbexploit':{"acq_func":"LCB","kappa":1.0},
                             'pi':{"acq_func":"pi"},
                             'xi02':{"xi":0.02},
                             'default':{},
                             'sampling':{'acq_optimizer':'sampling'},
                             'lbfgs':{'acq_optimizer':'lbfgs'},
                             'noisy':{'noise':'gaussian'},
                             'calm':{'noise':1e-8},
                             '':{}
                           })
# This could be improved I'm sure 
# https://github.com/scikit-optimize/scikit-optimize/discussions/1002


def skopt_gp_method_to_kwargs(method:str)->dict:
    """ Create param list for skopt.gp_minimize from name of strategy
         e.g.  lcbexplore_sampling_sobol -> {'acq_func':'LCB',... etc}
    """
    cues = method.split('_')
    params = dict()
    for cue in cues:
        if cue:
            if GP_INTERPRETATIONS.get(cue):
                params.update(deepcopy(GP_INTERPRETATIONS[cue]))
            else:
                msg = 'skopt_gp cannot interpret the cue ' + cue
                raise ValueError(msg)
    return params


def skopt_gp_cube_factory(objective, n_trials, n_dim, with_count=False, method=''):
    bounds = [(0.,1.,'uniform') ]*n_dim  # <-- Beware: rookie error to accidentally use corners only
                                         # when the dimension isn't explicitly stated as continuous

    global feval_count
    feval_count = 0

    def _objective(x):
        global feval_count
        feval_count +=1
        return objective(list(x))

    gp_params = skopt_gp_method_to_kwargs(method=method)
    result = gp_minimize(func=_objective,  dimensions=bounds, n_calls=n_trials, n_jobs=1, **gp_params)
    best_x = list(result.x)
    best_val = result.fun
    return (best_val, best_x,  feval_count) if with_count else (best_val, best_x)


# We could now create the following functions programmatically, but it is somewhat easier to debug this way

def skopt_gp_default_cube(objective, n_trials, n_dim, with_count=False):
    return skopt_gp_cube_factory(objective=objective,n_trials=n_trials,n_dim=n_dim, with_count=with_count,
                                 method='')


def skopt_gp_sobol_cube(objective, n_trials, n_dim, with_count=False):
        return skopt_gp_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                     method='sobol')


def skopt_gp_hammersley_cube(objective, n_trials, n_dim, with_count=False):
    return skopt_gp_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='hammersly')


def skopt_gp_lcb_cube(objective, n_trials, n_dim, with_count=False):
    return skopt_gp_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='lcb')


def skopt_gp_lcbexplore_cube(objective, n_trials, n_dim, with_count=False):
    return skopt_gp_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='lcbexplore')


def skopt_gp_lcbexploit_cube(objective, n_trials, n_dim, with_count=False):
    return skopt_gp_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='lcbexploit')


def skopt_gp_sampling_cube(objective, n_trials, n_dim, with_count=False):
    return skopt_gp_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='sampling')


SKOPT_GP_CANDIDATES = [ skopt_gp_default_cube, skopt_gp_sobol_cube, skopt_gp_hammersley_cube,
                        skopt_gp_lcb_cube, skopt_gp_lcbexplore_cube, skopt_gp_lcbexploit_cube,
                        skopt_gp_sampling_cube ]

SKOPT_GP_OPTIMIZERS = [ skopt_gp_default_cube ]

if __name__ == '__main__':
    always_working = SKOPT_GP_CANDIDATES[:1]
    broken = set()
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in SKOPT_GP_CANDIDATES:
            try:
                print(optimizer(objective, n_trials=250, n_dim=6, with_count=True))
            except Exception as e:
                print(e)
                broken.add(optimizer)
                always_working.remove(optimizer)
    print(' ')
    print('Sometimes broken: ')
    print([b.__name__ for b in broken])
    print(' ')
    print('Alway working: ')
    print([b.__name__ for b in always_working])






import numpy as np
from timemachines.common.plotting import mesh2d
from timemachines.objectives.planar_h1 import hump_selector
from shgo import shgo
from sklearn.datasets import make_spd_matrix

# Just messin' around here.
# Moves h1-like hump functions around on the plane until
# finding the minima of them because easier.


def locations_objective(zs, random_ws, offset):
    """ Return mean minimum found
        zs - center of humps
        ws - weights
        offset - offset to apply to hump function
    """
    scores = list()


    for ws_ in random_ws:
        def hs(xs):
            return hump_selector(xs=xs,zs=zs,ws=ws_,offset=offset)
        dim = 2
        bounds = [(-0.5,0.5)]*dim
        res = shgo(hs, bounds, n=10, iters=2)
        score = 100*res.fun + res.nfev
        scores.append(score)
    return np.mean(scores)


def best_locations(ws,offset, random_ws):
    """
        For given weights on humps, move them around until it helps the search

        z0   Initial locations of humps in flat format [ x1, y1, x2, y2, ... ]
        ws   Weights assigned to each hump
    """

    # Simulate random hump multipliers


    def mb(zs):
        return locations_objective(zs=np.array(zs), random_ws=random_ws, offset=offset)

    dim = int(len(ws))
    bounds = [(-0.5,0.5)]*dim*2
    res = shgo(mb, bounds, n=100, iters=50)
    print(res.fun)
    return res.x


if __name__=='__main__':

    ws = np.array([0.5,0.5,0.5])

    dim = len(ws)
    zs = 0.333*np.random.rand(2*dim)
    offset = 0.1 * np.random.randn(2)
    zdim = len(ws)
    num_ws = 10

    C = make_spd_matrix(zdim)
    while (C[0][1])**2/(C[0][0]*C[1][1])<0.97:
        C = make_spd_matrix(zdim)
    print(C)
    random_ws = np.random.multivariate_normal(ws, 2*C, num_ws)

    avg = locations_objective(zs=zs, random_ws=random_ws, offset=offset)
    print(avg)

    # Optimize positions of humps making it easier
    zBest = best_locations(ws=ws,offset=offset,random_ws=random_ws)
    print(zBest)
    print(ws)
    import random
    f = lambda x1, x2: hump_selector(xs=np.array([x1, x2]), zs=zBest, ws=random.choice(random_ws), offset=offset)
    mesh2d(f)
    pass






from bayes_opt import BayesianOptimization
from humpday.objectives.classic import CLASSIC_OBJECTIVES


BAYESOPT_METHODS = ['ucb', 'ei', 'poi']

def bayesopt_cube_factory(objective, n_trials, n_dim, with_count,method):

    assert method in BAYESOPT_METHODS

    global feval_count
    feval_count = 0

    pbounds = dict([('u' + str(i), (0., 1.)) for i in range(n_dim)])

    def _neg_objective(**kwargs) -> float:
        global feval_count
        feval_count += 1
        u = [kwargs['u' + str(i)] for i in range(n_dim)]
        return -objective(u)

    optimizer = BayesianOptimization(
        f=_neg_objective,
        pbounds=pbounds,
        verbose=0,
        random_state=1,
    )

    optimizer.maximize(
        init_points=5,
        acq=method,
        n_iter=n_trials - 5,
    )

    best_val = -optimizer.max['target']
    best_x = [optimizer.max['params']['u' + str(i)] for i in range(n_dim)]

    return (best_val, best_x, feval_count) if with_count else (best_val, best_x)


def bayesopt_ucb_cube(objective, n_trials, n_dim, with_count):
    return bayesopt_cube_factory(objective=objective,n_trials=n_trials,
                                 n_dim=n_dim, with_count=with_count,method='ucb')


def bayesopt_ei_cube(objective, n_trials, n_dim, with_count):
    return bayesopt_cube_factory(objective=objective,n_trials=n_trials,
                                 n_dim=n_dim, with_count=with_count,method='ei')


def bayesopt_poi_cube(objective, n_trials, n_dim, with_count):
    return bayesopt_cube_factory(objective=objective,n_trials=n_trials,
                                 n_dim=n_dim, with_count=with_count,method='poi')


BAYESOPT_OPTIMIZERS = [ bayesopt_ucb_cube, bayesopt_ei_cube, bayesopt_poi_cube ]


if __name__=='__main__':
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in BAYESOPT_OPTIMIZERS:
            print((optimizer(objective, n_trials=25, n_dim=3, with_count=True)))


from poap.controller import BasicWorkerThread, ThreadController
from pySOT.experimental_design import SymmetricLatinHypercube, LatinHypercube, TwoFactorial
from pySOT.strategy import SRBFStrategy, EIStrategy, DYCORSStrategy,RandomStrategy, LCBStrategy
from pySOT.surrogate import CubicKernel, LinearTail, RBFInterpolant, GPRegressor
from pySOT.optimization_problems.optimization_problem import OptimizationProblem
import numpy as np
import logging

logging.getLogger('pySOT').setLevel(logging.ERROR)


class GenericProblem(OptimizationProblem):

    def __init__(self, dim, objective):
        self.dim = dim
        self.min = 0
        self.minimum = np.zeros(dim)
        self.lb = 0 * np.ones(dim)
        self.ub = 1 * np.ones(dim)
        self.int_var = np.array([])
        self.cont_var = np.arange(0, dim)
        self.objective = objective
        self.info = str(dim) + "-dimensional objective function " + objective.__name__
        self.feval_count = 0

    def eval(self, x):
        """Evaluate the objective x

        :param x: Data point
        :type x: numpy.array
        :return: Value at x
        :rtype: float
        """
        self.__check_input__(x)
        d = float(self.dim)
        self.feval_count = self.feval_count+1
        return self.objective(list(x))


def pysot_cube( objective, n_trials, n_dim, with_count=False, method=None, design=None):
    """ Minimize
    :param objective:
    :param n_trials:
    :param n_dim:
    :param with_count:
    :return:
    """
    logging.getLogger('pySOT').setLevel(logging.ERROR)

    num_threads = 1
    asynchronous = True

    max_evals = n_trials
    gp = GenericProblem(dim=n_dim, objective=objective)

    if design=='latin':
        exp_design = LatinHypercube(dim=n_dim, num_pts=2 * (n_dim + 1))
    elif design=='symmetric':
        exp_design = SymmetricLatinHypercube(dim=n_dim, num_pts=2 * (n_dim + 1))
    elif design=='factorial':
        exp_design = TwoFactorial(dim=n_dim)
    else:
        raise ValueError('design should be latin, symmetric or factorial')

    # Create a strategy and a controller
    #  SRBFStrategy, EIStrategy, DYCORSStrategy,RandomStrategy, LCBStrategy
    controller = ThreadController()
    if method.lower()=='srbf':
        surrogate = RBFInterpolant(dim=n_dim, lb=np.array([0.0] * n_dim), ub=np.array([1.0] * n_dim), kernel=CubicKernel(),
                             tail=LinearTail(n_dim))
        controller.strategy = SRBFStrategy(
            max_evals=max_evals, opt_prob=gp, exp_design=exp_design, surrogate=surrogate, asynchronous=asynchronous
        )
    elif method.lower() == 'ei':
        surrogate = GPRegressor(dim=n_dim, lb=np.array([0.0] * n_dim), ub=np.array([1.0] * n_dim) )
        controller.strategy = EIStrategy(
            max_evals=max_evals, opt_prob=gp, exp_design=exp_design, surrogate=surrogate, asynchronous=asynchronous
        )
    elif method.lower() == 'dycors':
        surrogate = RBFInterpolant(dim=n_dim, lb=np.array([0.0] * n_dim), ub=np.array([1.0] * n_dim), kernel=CubicKernel(),
                             tail=LinearTail(n_dim))
        controller.strategy = DYCORSStrategy(
            max_evals=max_evals, opt_prob=gp, exp_design=exp_design, surrogate=surrogate, asynchronous=asynchronous
        )
    elif method.lower() == 'lcb':
        surrogate = GPRegressor(dim=n_dim, lb=np.array([0.0] * n_dim), ub=np.array([1.0] * n_dim))
        controller.strategy = LCBStrategy(
            max_evals=max_evals, opt_prob=gp, exp_design=exp_design, surrogate=surrogate, asynchronous=asynchronous
        )
    elif method.lower() == 'random':
        controller.strategy = RandomStrategy(
            max_evals=max_evals, opt_prob=gp
        )
    else:
        raise ValueError("Didn't recognize method passed to pysot")


    # Launch the threads and give them access to the objective function
    for _ in range(num_threads):
        worker = BasicWorkerThread(controller, gp.eval)
        controller.launch_worker(worker)

    # Run the optimization strategy
    result = controller.run()
    best_x = result.params[0].tolist()
    return (result.value, best_x, gp.feval_count) if with_count else (result.value, best_x)


# Index by strategy
# SRBFStrategy, EIStrategy, DYCORSStrategy,RandomStrategy, LCBStrategy

def pysot_srbf_cube( objective, n_trials, n_dim, with_count=False):
    return pysot_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='srbf', design='symmetric' )


def pysot_ei_cube( objective, n_trials, n_dim, with_count=False):
    return pysot_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='ei', design='symmetric' )


def pysot_dycors_cube( objective, n_trials, n_dim, with_count=False):
    return pysot_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='dycors', design='symmetric' )


def pysot_lcb_cube( objective, n_trials, n_dim, with_count=False):
    return pysot_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='lcb', design='symmetric' )


def pysot_random_cube( objective, n_trials, n_dim, with_count=False):
    return pysot_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='random', design='symmetric' )


PYSOT_OPTIMIZERS = [ pysot_ei_cube, pysot_lcb_cube, pysot_random_cube, pysot_srbf_cube, pysot_dycors_cube ]


if __name__ == '__main__':
    from humpday.objectives.classic import CLASSIC_OBJECTIVES
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in PYSOT_OPTIMIZERS:
            print( (optimizer.__name__, optimizer(objective, n_trials=250, n_dim=6, with_count=True)))


from ax import optimize
from logging import CRITICAL
from ax.utils.common.logger import get_logger
rt = get_logger('ax')
rt.setLevel(CRITICAL)
import warnings
warnings.filterwarnings("ignore")
from funcy import print_durations
from humpday.objectives.classic import CLASSIC_OBJECTIVES


def ax_cube(objective, n_trials, n_dim, with_count=False, method=None):
    global feval_count
    feval_count = 0

    def evaluation_func(prms):
        global feval_count
        feval_count += 1
        return objective([prms["u" + str(i)] for i in range(n_dim)])

    parameters = [{
        "name": "u" + str(i),
        "type": "range",
        "bounds": [0.0, 1.0],
    } for i in range(n_dim)]
    best_parameters, best_values, experiment, model = optimize(parameters=parameters,
                                                               evaluation_function=evaluation_func,
                                                               minimize=True,
                                                               total_trials=n_trials)
    best_x = [ best_parameters['u'+str(i)] for i in range(n_dim) ]
    best_val = best_values[0]['objective']
    return (best_val, best_x, feval_count) if with_count else (best_val, best_x)


def ax_default_cube(objective, n_trials, n_dim, with_count=False):
    return ax_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count)


AX_OPTIMIZERS = [ ax_default_cube ]

@print_durations()
def demo():
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in AX_OPTIMIZERS:
            print(optimizer(objective, n_trials=250, n_dim=6, with_count=True))


if __name__=='__main__':
    demo()


from humpday.optimizers.alloptimizers import OPTIMIZERS
from humpday.objectives.classic import CLASSIC_OBJECTIVES
from pprint import pprint
from typing import List
from humpday.comparison.naming import optimizer_name

# See also https://github.com/microprediction/humpday-testing/tree/main/optimizer_elo_ratings/leaderboards


def comparison(objectives:List=None, n_dim=5, n_trials=100):
    """  Run all optimizers and return ranked list
    :param objectives:  A list of functions taking a single argument representing [0,1]^n_dim
    :param n_trials:    Maximum number of function evaluations
    :returns sorted list of minima found and actual number of function evaluations
    """
    if objectives is None:
        objectives = CLASSIC_OBJECTIVES
    return sorted([(optimizer(objective, n_trials=n_trials, n_dim=n_dim, with_count=True),
                          optimizer_name(optimizer),objective.__name__)
                         for optimizer in OPTIMIZERS for objective in objectives])


if __name__=='__main__':
    pprint( comparison() )







from deap import benchmarks
import numpy as np
import math

# Some test objective functions to help guide optimizer choices
# -------------------------------------------------------------
#
# We'll use DEAP's set of groovy benchmarks, and landscapes, swarmpackagepy also
#
# See pretty pictures at https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks
# Some hardness assessment is at https://github.com/nathanrooy/landscapes#available-functions-from-single_objective but we'll do our own


## Basis of tricky functions

import datetime
DAY = datetime.datetime.today().day
OFFSET = DAY/50
POWER = 1+ (DAY % 3)/3.0
SHIFT = DAY/100


def smoosh(ui):
    """ Distort the interval to avoid obvious minima and avoid memorization """
    ui_rotate = ui+SHIFT % 1.0
    ui_shift = ui_rotate + SHIFT
    xi   = ui_shift**POWER
    low  = SHIFT**POWER
    high = (1+SHIFT)**POWER
    yi   = (xi-low)/(high-low)
    return yi**POWER


def schwefel_on_cube(u:[float])->float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.schwefel
    u_squished = [ 1000*(smoosh(ui)-0.5) for ui in u ]
    try:
        return 0.001*benchmarks.schwefel(u_squished)[0]/0.71063
    except Exception as e:
        raise Exception(e)


def griewank_on_cube(u:[float])->float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.griewank
    u_squished = [ 1200*(ui**1.1-0.5) for ui in u ]
    return benchmarks.griewank(u_squished)[0]/0.532075


def rastrigin_on_cube(u:[float])->float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.rastrigin
    u_squished = [ 10.24*(ui**1.1-0.5) for ui in u ]
    return 0.01*benchmarks.rastrigin(u_squished)[0]/0.059697


def bohachevsky_on_cube(u:[float])->float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.bohachevsky
    u_squished = [ 10*(ui**1.1-0.5) for ui in u ]
    return 1.0+benchmarks.bohachevsky(u_squished)[0]


def rosenbrock_on_cube(u:[float])->float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.rosenbrock
    u_squished = [ 200*(ui**1.1-0.5) for ui in u ]
    return 1+0.1*benchmarks.rosenbrock(u_squished)[0]/0.008949


def shaffer_on_cube(u:[float])->float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.schaffer
    u_squished = [ 200*(ui**1.1-0.5) for ui in u ]
    return 0.01*benchmarks.schaffer(u_squished)[0]/(0.1042133*0.71809)


def shekel_on_cube(u:[float])->float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.schaffer

    n_dim = len(u)
    NUMMAX = 15
    A = 10 * np.random.rand(NUMMAX, n_dim)
    C = np.random.rand(NUMMAX)
    u_squished = [ 800*(smoosh(ui)-0.5) for ui in u ]
    return 1.2298-benchmarks.shekel(u_squished,A,C)[0]


## Combinations

def deap_combo1_on_cube(u:[float])->float:
    return 0.3*(schwefel_on_cube(u) + griewank_on_cube(u)+shekel_on_cube(u))/1.883


def deap_combo2_on_cube(u:[float])->float:
    return 0.5*(shaffer_on_cube(u) + shekel_on_cube(u))-0.1075


def deap_combo3_on_cube(u:[float])->float:
    return 0.5*(rosenbrock_on_cube(u) + bohachevsky_on_cube(u)+shekel_on_cube(u))/1.88


DEAP_OBJECTIVES = [schwefel_on_cube, rastrigin_on_cube, griewank_on_cube,
                   bohachevsky_on_cube, rosenbrock_on_cube, shaffer_on_cube, shekel_on_cube,
                   deap_combo1_on_cube, deap_combo2_on_cube, deap_combo3_on_cube]


# By hand...

def rosenbrock_modified_on_cube(u:[float])->float:
    """ https://en.wikipedia.org/wiki/Rosenbrock_function """
    u_scaled = [4 * ui - 2 for ui in u]
    if len(u)==1:
        return (0.25-u_scaled[0])**2
    else:
        return 5+0.001*np.sum([ 100*(ui_plus-ui*ui)+(1-ui)*(1-ui) for ui,ui_plus in zip(u_scaled[1:],u_scaled)])


# According to http://infinity77.net/global_optimization/test_functions.html#test-functions-index
# there are some really hard ones
# See https://github.com/andyfaff/ampgo/blob/master/%20ampgo%20--username%20andrea.gavana%40gmail.com/go_benchmark.py
# See also https://arxiv.org/pdf/1308.4008v1.pdf

def damavandi_on_cube(u:[float])->float:
    """ A trivial multi-dimensional extension of Damavandi's function """
    return 0.01*damavandi2(u[0],u[1])-0.46


def damavandi2(u1,u2)->float:
    """ Pretty evil function this one """
    # http://infinity77.net/global_optimization/test_functions_nd_D.html#go_benchmark.Damavandi
    x1 = u1/14.
    x2 = u2/14.
    numerator = math.sin(math.pi*(x1 - 2.0))*math.sin(math.pi*(x2 - 2.0))
    denumerator = (math.pi**2)*(x1 - 2.0)*(x2 - 2.0)
    factor1 = 1.0 - (abs(numerator / denumerator))**5.0
    factor2 = 2 + (x1 - 7.0)**2.0 + 2*(x2 - 7.0)**2.0
    return factor1*factor2


def paviani_on_cube(u:[float])->float:
    # http://infinity77.net/global_optimization/test_functions_nd_P.html#go_benchmark.Paviani
    x = np.array([ 2.001 + 5.996*smoosh(ui) for ui in u])

    def safe_np_log(x):
        lb = np.array([1e-6]*len(x))
        xup = np.maximum(x,lb)
        return np.log(xup)

    return float( np.sum(safe_np_log(x-2)**2.0 + safe_np_log(10.0 - x)**2.0) - np.prod(x)**0.2 )/8.6456


# Landscapes

from landscapes.single_objective import styblinski_tang, zakharov, salomon, rotated_hyper_ellipsoid, qing, michalewicz


def styblinski_tang_on_cube(u:[float])->float:
    u_scaled = [10*(smoosh(ui)-0.5) for ui in u]
    return 3.3499+0.01*styblinski_tang(u_scaled)


def zakharov_on_cube(u:[float])->float:
    u_scaled = [15*smoosh(ui)-10 for ui in u]
    return 0.01*zakharov(u_scaled)/0.3462


def salomon_on_cube(u:[float])->float:
    u_scaled = [200*smoosh(ui) - 100 for ui in u]
    return salomon(u_scaled)/3.09999


def rotated_hyper_ellipsoid_on_cube(u:[float])->float:
    u_scaled = [2 * 65.536*smoosh(ui) - 65.536 for ui in u]
    return 0.1*rotated_hyper_ellipsoid(u_scaled)


def qing_on_cube(u:[float])->float:
    u_scaled = [1000*smoosh(ui) - 500 for ui in u]
    return qing(u_scaled)/0.01805


def michaelewicz_on_cube(u:[float])->float:
    u_scaled = [4*smoosh(ui) - 2 for ui in u]
    return 1.4439 + 0.1*michalewicz(u_scaled,m=20)


def landscapes_combo1_on_cube(u:[float])->float:
    return (qing_on_cube(u) + michaelewicz_on_cube(u))/(1.5744*1.4688)


def landscapes_combo2_on_cube(u:[float])->float:
    return (rotated_hyper_ellipsoid_on_cube(u) + salomon_on_cube(u))/(6.7555*0.82)


def landscapes_combo3_on_cube(u:[float])->float:
    return (2 + zakharov_on_cube(u) + styblinski_tang_on_cube(u))/4.4329


LANDSCAPES_OBJECTIVES = [ styblinski_tang_on_cube, zakharov_on_cube, salomon_on_cube, rotated_hyper_ellipsoid_on_cube,
                          qing_on_cube, michaelewicz_on_cube, landscapes_combo1_on_cube, landscapes_combo2_on_cube,
                          landscapes_combo3_on_cube ]

# Some copied from peabox
# https://github.com/stromatolith/peabox/blob/master/peabox/peabox_testfuncs.py
# as that isn't deployed to PyPI as far as I can determine


def ackley_on_cube(u:[float])->float:
    # allow parameter range -32.768<=x(i)<=32.768, global minimum at x=(0,0,...,0)
    rescaled_u = [2*32.768*smoosh(ui) - 32.768 for ui in u]
    x = np.asfarray(rescaled_u)
    ndim=len(x)
    a=20.; b=0.2; c=2.*math.pi
    return (-a*np.exp(-b*np.sqrt(1./ndim*np.sum(x**2)))-np.exp(1./ndim*np.sum(np.cos(c*x)))+a+np.exp(1.))/20.0


# Adapted from https://github.com/SISDevelop/SwarmPackagePy/blob/master/SwarmPackagePy/testFunctions.py

def cross_on_cube(u):
    x = [5 * smoosh(ui) - 2.5 for ui in u]
    return round(-0.0001*(abs(math.sin(x[0])*math.sin(x[1])*math.exp(abs(100 -
                            math.sqrt(sum([i**2 for i in x]))/math.pi))) + 1)**0.1, 7)


def powers_on_cube(u):
    x = [5 * smoosh(ui) - 2.5 for ui in u]
    return sum([abs(x[i])**(i+2) for i in range(len(x))])


def booth_on_cube(u):
    x = [5 * smoosh(ui) - 2.5 for ui in u]
    return sum([abs(x[i])**(i+2) for i in range(len(x))])


def matyas_on_cube(u):
    x = [3 * smoosh(ui) - 1.5 for ui in u]

    def sphere_function(x):
        return sum([i ** 2 for i in x])

    return 0.26*sphere_function(x) - 0.48*x[0]*x[1]


def drop_wave_on_cube(u):
    x = [3 * smoosh(ui) - 1.5 for ui in u]

    def sphere_function(x):
        return sum([i ** 2 for i in x])

    return -(1 + math.cos(12*math.sqrt(sphere_function(x))))/(0.5*sphere_function(x) + 2)

SWARM_OBJECTIVES = [cross_on_cube, powers_on_cube,
                    booth_on_cube, matyas_on_cube, drop_wave_on_cube]


A_CLASSIC_OBJECTIVE = rastrigin_on_cube  # Just pick one for testing

MISC_OBJECTIVES = [ paviani_on_cube, damavandi_on_cube, rosenbrock_modified_on_cube, ackley_on_cube ]

CLASSIC_OBJECTIVES = DEAP_OBJECTIVES + LANDSCAPES_OBJECTIVES + MISC_OBJECTIVES + SWARM_OBJECTIVES


if __name__=="__main__":
    for objective in CLASSIC_OBJECTIVES:
        objective(u=[0.0,0.5,1.0])
        objective(u=[0.0, 0.5, 0.0, 0.0, 1.0])
    print(len(CLASSIC_OBJECTIVES))



from humpday.optimizers.alloptimizers import OPTIMIZERS
from humpday.objectives.classic import CLASSIC_OBJECTIVES
from humpday.comparison.naming import optimizer_name

# TODO: Add speed check


def whats_broken():
    for n_trials in [2,10,40]:
        for n_dim in [2,3,8]:
            for optimizer in OPTIMIZERS:
                for objective in CLASSIC_OBJECTIVES:
                    try:
                        optimizer(objective, n_trials=n_trials,n_dim=n_dim,with_count=True)
                    except Exception as e:
                        print(e)
                        print(optimizer_name(optimizer) + ' fails on ' + objective.__name__ +
                              ' in dimension '+str(n_dim)+' when n_trials='+str(n_trials))


if __name__=='__main__':
    whats_broken()


#!/usr/bin/env python

# add command line args, clean up output a little, add timings, collect results into table
from humpday.optimizers.hyperoptcube import HYPEROPT_OPTIMIZERS
from humpday.optimizers.shgocube import SHGO_OPTIMIZERS
from humpday.optimizers.optunacube import OPTUNA_OPTIMIZERS
from humpday.optimizers.pysotcube import PYSOT_OPTIMIZERS
from humpday.optimizers.scipycube import SCIPY_OPTIMIZERS
from humpday.optimizers.axcube import AX_OPTIMIZERS
from humpday.optimizers.platypuscube import PLATYPUS_OPTIMIZERS
from humpday.optimizers.pymoocube import PYMOO_OPTMIZERS
from humpday.optimizers.swarmlibcube import SWARMLIB_OPTIZERS
from humpday.optimizers.nevergradcube import NEVERGRAD_OPTIMIZERS
from humpday.objectives.classic import CLASSIC_OBJECTIVES
from humpday.optimizers.skoptcube import SKOPT_GP_OPTIMIZERS
from humpday.optimizers.ultraoptcube import ULTRAOPT_OPTIMIZERS
from humpday.optimizers.bayesoptcube import BAYESOPT_OPTIMIZERS
from humpday.optimizers.dlibcube import DLIB_OPTIMIZERS

from datetime import datetime
import pandas as pd
import argparse

CANDIDATES = SCIPY_OPTIMIZERS + SHGO_OPTIMIZERS + HYPEROPT_OPTIMIZERS +\
             PYSOT_OPTIMIZERS + OPTUNA_OPTIMIZERS + AX_OPTIMIZERS +\
             PLATYPUS_OPTIMIZERS + PYMOO_OPTMIZERS + NEVERGRAD_OPTIMIZERS + SWARMLIB_OPTIZERS + SKOPT_GP_OPTIMIZERS

# To see what might be working, or not, refer regression testing results in directories such as:
# https://github.com/microprediction/humpday-testing/tree/main/data/brownian/dlm_seasonal
OPTIMIZERS = SHGO_OPTIMIZERS + SCIPY_OPTIMIZERS + PYSOT_OPTIMIZERS + AX_OPTIMIZERS + \
             OPTUNA_OPTIMIZERS + PLATYPUS_OPTIMIZERS + NEVERGRAD_OPTIMIZERS +\
             SWARMLIB_OPTIZERS + HYPEROPT_OPTIMIZERS + PYMOO_OPTMIZERS +\
             SKOPT_GP_OPTIMIZERS + ULTRAOPT_OPTIMIZERS + BAYESOPT_OPTIMIZERS + DLIB_OPTIMIZERS


def optimizer_from_name(name):
    valid = [f for f in OPTIMIZERS if f.__name__==name ]
    return valid[0] if len(valid)==1 else None


if __name__=='__main__':

    parser = argparse.ArgumentParser(description="Run all optimizers on input size ndim (default 2) requesting ntrials (default 20) iterations and save results (default log.csv")

    parser.add_argument("-d", "--ndims", type=int, action="extend", nargs="+",
                        help="Number of input dimensions to objective function (default 2)")
    parser.add_argument("-t", "--ntrials", type=int, action="extend", nargs="+",
                        help="Number of trial iterations in optimization (default 20)")
    parser.add_argument("-v", "--verbose", help="Increase output verbosity",
                        action="store_true")
    parser.add_argument("-o", "--logfile", help="Specify outputfile (default log.csv)")
    args = parser.parse_args()
    # print(args)
    LOGFILE = 'log.csv'
    if args.logfile is not None:
        LOGFILE = args.logfile

    NDIMS = [2]
    if args.ndims is not None:
        NDIMS = args.ndims
        
    NTRIALS = [20]
    if args.ntrials is not None:
        NTRIALS = args.ntrials

    print(' ')
    print('Full list of optimizer strategies .. ')
    print(' ')
    print([ o.__name__.replace('_cube','') for o in OPTIMIZERS])
    print(' ')
    print('Full list of objective functions .. ')
    print(' ')
    print([o.__name__ for o in CLASSIC_OBJECTIVES])
    print(' ')

    print(datetime.now(), str(len(OPTIMIZERS)) + ' optimization strategies will be compared.')
    print(datetime.now(), str(len(CLASSIC_OBJECTIVES)) + ' objective functions will be employed.')
    print(datetime.now(), 'objective input dimensions: ', str(NDIMS))
    print(datetime.now(), 'number of trials: ', str(NTRIALS))
    print(datetime.now(), 'logfile: ', LOGFILE)

    log_array = []
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        for n_dim in NDIMS:
            for n_trials in NTRIALS:
                print(datetime.now(), 'Now testing against '+objective.__name__+' in '+str(n_dim)+' dimensions requesting '+str(n_trials)+' trials.')
                for optimizer in OPTIMIZERS:
                    try:
                        start_time = datetime.now()
                        result = optimizer(objective, n_trials=n_trials, n_dim=n_dim, with_count=True)
                        best_value, best_params, reported_trials = result
                        end_time = datetime.now()
                        time_elapsed = datetime.now() - start_time
                        print(datetime.now(), "Finished", reported_trials, "trials in ", time_elapsed)
                        print(datetime.now(), optimizer.__name__, result)
                        log_array.append([start_time, end_time, time_elapsed,
                                          objective.__name__, optimizer.__name__,
                                          n_dim, n_trials, reported_trials,
                                          best_value, best_params, ])
                    except:
                        import warnings
                        print(' ')
                        warnings.warn(' WARNING : '+optimizer.__name__+' fails on '+ objective.__name__+ ' in '+str(n_dim)+' dimensions with '+str(n_trials)+' trials.')

    log_df = pd.DataFrame(log_array)
    log_df.columns = ["start_time", "end_time", "time_elapsed",
                      "objective", "optimizer",
                      "n_dim", "n_trials", "reported_trials",
                      "best_value", "best_params", ]
    print(log_df)
    log_df.to_csv(LOGFILE, index=False)



import warnings

# http://dlib.net/optimization.html#find_min_global
# This library also provides global_function_search which is pretty darn cool

try:
    from dlib import find_min_global
    using_dlib = True
except ImportError:
    warnings.warn('dlib may not be properly supported on this operating system. Maybe try pip install --upgrade cmake')
    using_dlib = False

if using_dlib:

    def dlib_cube(objective ,n_trials, n_dim, with_count):
        global feval_count
        feval_count = 0

        lb = [0. for _ in range(n_dim)]
        ub = [1. for _ in range(n_dim)]

        def _objective(*args) -> float:
            global feval_count
            feval_count += 1
            return objective(list(args))

        best_x, best_val = find_min_global(_objective, lb, ub, n_trials)

        return (best_val, best_x, feval_count) if with_count else (best_val, best_x)


    DLIB_OPTIMIZERS = [dlib_cube]
else:
    DLIB_OPTIMIZERS = []


if __name__ == '__main__':
    from humpday.objectives.classic import CLASSIC_OBJECTIVES

    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in DLIB_OPTIMIZERS:
            print(optimizer(objective, n_trials=500, n_dim=34, with_count=True))


from platypus import NSGAII, Problem, Real, EvolutionaryStrategy, GeneticAlgorithm, \
    NSGAIII, CMAES, GDE3, IBEA, MOEAD, OMOPSO, SMPSO, SPEA2, EpsMOEA, normal_boundary_weights
from platypus.core import FixedLengthArray
import logging

logging.getLogger('Platypus').setLevel(logging.ERROR)

# Platypus is designed for multi-objective optimization, so may not do so well on single-objective problems

PLATYPUS_ALGORITHMS = {'evolutionary': EvolutionaryStrategy,
                       'genetic': GeneticAlgorithm,
                       'nsgaii': NSGAII,
                       'nsgaiii': (NSGAIII, {"divisions_outer": 12}),
                       'cmaes': (CMAES, {"epsilons": [0.05]}),
                       'gde3': GDE3,
                       'ibea': IBEA,
                       'moead': (MOEAD, {"weight_generator": normal_boundary_weights, "divisions_outer": 12}),
                       'omopso': (OMOPSO, {"epsilons": [0.05]}),
                       'smpso': SMPSO,
                       'spea2': SPEA2,
                       'epsmoea': (EpsMOEA, {"epsilons": [0.05]})
                       }


def platypus_cube(objective, n_trials, n_dim, with_count=False, method=None):
    global feval_count
    feval_count = 0

    def _objective(vars):
        global feval_count
        feval_count += 1
        return float(objective(list(vars)))  # Avoid np.array as Platypus may puke

    problem = Problem(n_dim, 1, 0)
    problem.types[:] = [Real(0.0, 1.0)] * n_dim
    problem.constraints[:] = "<=0"
    problem.function = _objective

    strategy_and_args = PLATYPUS_ALGORITHMS[method]
    if isinstance(strategy_and_args, tuple):
        strategy = strategy_and_args[0]
        strategy_args = strategy_and_args[1]
        algorithm = strategy(problem, **strategy_args)
    else:
        strategy = strategy_and_args
        algorithm = strategy(problem)

    algorithm.run(n_trials)
    feasible_solution_obj = sorted([(s.objectives[0], s.variables) for s in algorithm.result if s.feasible],
                                   reverse=False)
    best_obj, best_x = feasible_solution_obj[0]
    if isinstance(best_x, FixedLengthArray):
        best_x = best_x._data  # CMA-ES returns it this way for some reason
    return (best_obj, best_x, feval_count) if with_count else (best_obj, best_x)


def platypus_evolutionary_cube(objective, n_trials, n_dim, with_count=False):
    return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                         method='genetic')


def platypus_genetic_cube(objective, n_trials, n_dim, with_count=False):
    return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                         method='genetic')


def platypus_nsgaii_cube(objective, n_trials, n_dim, with_count=False):
    return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                         method='nsgaii')


def platypus_cmaes_cube(objective, n_trials, n_dim, with_count=False):
    return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                         method='cmaes')


def platypus_gde3_cube(objective, n_trials, n_dim, with_count=False):
    return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                         method='gde3')


def platypus_ibea_cube(objective, n_trials, n_dim, with_count=False):
    return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                         method='ibea')


def platypus_moead_cube(objective, n_trials, n_dim, with_count=False):
    return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                         method='moead')


def platypus_omopso_cube(objective, n_trials, n_dim, with_count=False):
    return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                         method='omopso')


def platypus_smpso_cube(objective, n_trials, n_dim, with_count=False):
    return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                         method='smpso')


def platypus_spea2_cube(objective, n_trials, n_dim, with_count=False):
    return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                         method='spea2')


def platypus_epsmoea_cube(objective, n_trials, n_dim, with_count=False):
    return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                         method='epsmoea')


BAD  = [ platypus_nsgaii_cube ]

PLATYPUS_OPTIMIZERS = [platypus_genetic_cube, platypus_evolutionary_cube,
                       platypus_nsgaii_cube, platypus_cmaes_cube, platypus_gde3_cube,
                       platypus_ibea_cube, platypus_moead_cube, platypus_omopso_cube,
                       platypus_smpso_cube, platypus_spea2_cube, platypus_epsmoea_cube]

if __name__ == '__main__':
    from humpday.objectives.classic import CLASSIC_OBJECTIVES

    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in PLATYPUS_OPTIMIZERS:
            print((optimizer.__name__, optimizer(objective, n_trials=250, n_dim=6, with_count=True)))



import numpy as np
import random
from pprint import pprint
import traceback
from humpday.objectives.classic import CLASSIC_OBJECTIVES
from humpday.optimizers.alloptimizers import OPTIMIZERS
from humpday.comparison.eloformulas import elo_update
from humpday.comparison.naming import optimizer_name, objective_name

N_DIM_CHOICES = [1, 2, 3, 5, 8]
N_TRIALS_CHOICES = [130, 210, 340]

OPTIMIZER_F_FACTOR = 1000
OPTIMIZER_K_FACTOR = 60
N_PROVISIONAL = 0  # Number of games for which player is considered provisional
N_ATTEMPTS_WHITE = 3
N_ATTEMPTS_BLACK = 6


def optimizer_game(white, black, n_dim, n_trials, objective, tol=0.001):
    """
    :param white:   optimizer
    :param black:   optimizer
    :param n_dim:
    :param n_trials:
    :param objective:
    :return:  dict
    """

    game_result = {'n_dim': n_dim, 'n_trials': n_trials, 'objective': objective.__name__, 'white': white,
                   'black': black,
                   'traceback': ['passing', 'passing'], 'best_val': [None, None], 'best_x': [None, None],
                   'feval_count': [None, None],
                   'n_trials_instructed': [None, None], 'passing': [None, None],
                   'completed': False}

    # White to play..
    n_white_trials = n_trials
    n_white_attempts = 0
    while True:
        try:
            white_best_val, white_best_x, white_feval_count = white(objective, n_trials=n_white_trials, n_dim=n_dim,
                                                                    with_count=True)
            white_passing, white_traceback = white_best_val is not None, 'passing'
        except Exception as e:
            white_traceback = traceback.format_exc()
            white_passing, white_best_x, white_best_val, white_feval_count = False, None, None, None
        if not white_passing or (white_feval_count <= 1.1*n_trials) or (n_white_attempts > N_ATTEMPTS_WHITE):
            break
        else:
            n_white_attempts += 1
            print('Playing white,' + white.__name__ + ' attempt ' + str(
                n_white_attempts + 1) + ' after instruction to use ' + str(n_white_trials) + ' resulted in ' + str(
                white_feval_count) + ' evaluations.')
            n_white_trials = int(0.7 * n_white_trials)
    game_result['best_val'][0] = white_best_val
    game_result['passing'][0] = white_passing
    white_success = white_passing and white_feval_count <= 1.1*n_trials
    if white_passing and n_white_trials > n_trials:
        white_traceback = 'White took ' + str(white_feval_count) + ' function evals when instructed to use ' + str(
            n_white_trials)
    game_result['traceback'][0] = white_traceback
    if white_passing:
        game_result['n_trials_instructed'][0] = n_white_trials
        game_result['feval_count'][0] = white_feval_count
        game_result['best_val'][0] = white_best_val
        game_result['best_x'][0] = white_best_x
    else:
        pass

    # Black to play
    if white_success:
        n_black_trials = white_feval_count  # <-- Tries to match the number of white actual evaluations
        n_black_attempts = 0
        while True:
            try:
                black_best_val, black_best_x, black_feval_count = black(objective, n_trials=n_black_trials,
                                                                        n_dim=n_dim,
                                                                        with_count=True)
                black_traceback, black_passing = 'passing', black_best_val is not None
            except Exception as e:
                black_traceback = traceback.format_exc()
                black_passing, black_best_x, black_best_val, black_feval_count = False, None, None, None

            if not black_passing or (black_feval_count <= 1.1*white_feval_count) or (
                    n_black_attempts > N_ATTEMPTS_BLACK):
                break
            else:
                n_black_attempts += 1
                print('Playing black, ' + black.__name__ + ' attempt ' + str(
                    n_black_attempts + 1) + ' after instruction to use ' + str(
                    n_black_trials) + ' resulted in ' + str(black_feval_count) + ' evaluations.')
                n_black_trials = int(0.85 * n_black_trials)
        black_success = black_passing and black_feval_count <= 1.1 * white_feval_count
        game_result['n_trials_instructed'][1] = n_black_trials
        game_result['feval_count'][1] = black_feval_count
        game_result['best_val'][1] = black_best_val
        if black_passing and n_black_trials > n_trials:
            black_traceback = 'Black took ' + str(black_feval_count) + ' function evals when instructed to use ' + str(
                n_black_trials)
        game_result['traceback'][1] = black_traceback
        game_result['passing'][1] = black_passing
        if black_passing:
            game_result['best_val'][1] = black_best_val
            game_result['best_x'][1] = black_best_x
            game_result['n_trials_instructed'][1] = n_black_trials
            game_result['feval_count'][1] = black_feval_count

    # Now that White and Black have both played...
    if white_success and black_success:
        game_result['completed'] = True
        small = tol * (abs(white_best_val) + abs(black_best_val))  # Ties
        points = 1. if white_best_val < black_best_val - small else 0. if black_best_val < white_best_val - small else 0.5
        game_result['points'] = points
        game_result['winner'] = white.__name__.replace('_cube',
                                                       '') if points > 0.75 else black.__name__ if points < 0.25 else 'draw'
        game_result['loser'] = black.__name__.replace('_cube',
                                                      '') if points > 0.75 else white.__name__ if points < 0.25 else 'draw'
        if game_result['winner'] != 'draw':
            print(game_result['winner'] + ' beats ' + game_result['loser'])
        else:
            print(black.__name__ + ' holds ' + white.__name__ + ' to a draw.')
    else:
        game_result['winner'] = 'incomplete'
        game_result['points'] = None

    return game_result


def random_optimizer_game(optimizers=None, objectives=None, n_dim_choices: [int] = None,
                          n_trials_choices: [int] = None, tol=0.001, announce=False):
    if n_dim_choices is None:
        n_dim_choices = N_DIM_CHOICES

    if n_trials_choices is None:
        n_trials_choices = N_TRIALS_CHOICES

    if objectives is None:
        objectives = CLASSIC_OBJECTIVES

    if optimizers is None:
        optimizers = OPTIMIZERS

    white, black = np.random.choice(optimizers, size=2, replace=False)
    matchup = {'n_dim':random.choice(n_dim_choices),
               'n_trials':random.choice(n_trials_choices),
               'white':white,
               'black':black,
               'objective': random.choice(objectives),
               'tol':tol}
    if announce:
        pprint(matchup)
    return optimizer_game(**matchup)


def optimizer_population_elo_update(optimizers, game_result: dict, elo: dict, initial_elo=1600, peg=True):
    """ Create or update elo ratings for optimizers

          optimizers - List of optimizers that were considered
          game_result - Produced by optimizer_game
          elo   - Dictionary containing the 'state' of the population (i.e. elo ratings and game counts)
          tol   - Objective function ratio that results in a tie being declared
          peg   - If True, optuna_random is set to initial rating every time

        Chooses random objective function, random dimensions and random number of trials
        Speed is not taken into account
    """

    if not elo:
        # Initialize game counts and Elo ratings
        elo['name'] = [f.__name__ for f in optimizers]
        elo['count'] = [0 for _ in optimizers]
        elo['rating'] = [initial_elo for _ in optimizers]
        elo['traceback'] = ['not yet run' for _ in optimizers]
        elo['active'] = [True for _ in optimizers]

    else:
        # Check for newcomers
        new_names = [f.__name__ for f in optimizers if f.__name__ not in elo['name']]
        for new_name in new_names:
            elo['name'].append(new_name)
            elo['count'].append(0)
            elo['rating'].append(initial_elo)
            elo['traceback'].append('not yet run')
            elo['active'].append(True)

    # Who is active?
    optimizer_names = [o.__name__ for o in optimizers]
    elo['active'] = [name_ in optimizer_names for name_ in elo['name']]

    # Peg to random sampling?
    if peg:
        elo['rating'] = [initial_elo if 'optuna_random' in name_ else r for r, name_ in zip(elo['rating'], elo['name'])]

    # Process results of match
    white_name = game_result['white'].__name__
    black_name = game_result['black'].__name__
    white_ndx = elo['name'].index(white_name)
    black_ndx = elo['name'].index(black_name)
    elo['traceback'][white_ndx] = game_result['traceback'][0]
    elo['traceback'][black_ndx] = game_result['traceback'][1]
    if game_result['completed']:
        elo['count'][white_ndx] += 1
        elo['count'][black_ndx] += 1
        points = game_result['points']
        print('>>>> ' + game_result['winner'])
        white_elo, black_elo = elo['rating'][white_ndx], elo['rating'][black_ndx]
        min_games = min(elo['count'][white_ndx], elo['count'][black_ndx])
        k = OPTIMIZER_K_FACTOR / 2.0 if min_games > 10 else OPTIMIZER_K_FACTOR
        new_white_elo, new_black_elo = elo_update(white_elo=white_elo, black_elo=black_elo, points=points, k=k,
                                                  f=OPTIMIZER_F_FACTOR)
        # Don't allow players with provisional ratings to impact other's.
        if elo['count'][black_ndx] >= N_PROVISIONAL:
            elo['rating'][white_ndx] = new_white_elo
        if elo['count'][white_ndx] > N_PROVISIONAL:
            elo['rating'][black_ndx] = new_black_elo
    else:
        print('>>>> incomplete ')

    return elo


def demo_optimizer_elo():
    # Run this to generate Elo ratings that will update for as long as you have the patience.
    elo = {}
    while True:
        game_result = random_optimizer_game(optimizers=OPTIMIZERS, objectives=CLASSIC_OBJECTIVES,
                                            n_dim_choices=N_DIM_CHOICES, n_trials_choices=N_TRIALS_CHOICES, tol=0.001)
        print(' Game...')
        pprint(game_result)

        elo = optimizer_population_elo_update(optimizers=OPTIMIZERS, elo=elo, game_result=game_result)
        if random.choice(list(range(5))) == 1:
            print(' ')
            pprint(sorted(list(zip(elo['rating'], elo['name'])), reverse=True))
            print(' ')


if __name__ == '__main__':
    demo_optimizer_elo()



from humpday.objectives.classic import CLASSIC_OBJECTIVES
import nevergrad as ng
import logging

logging.getLogger('nevergrad').setLevel(logging.ERROR)

# Facebook nevergrad https://facebookresearch.github.io/nevergrad/optimization.html


NEVER_ENOUGH = {'ngopt':'nGOpt is “meta”-optimizer which adapts to the provided settings (budget, number of workers, parametrization) and should therefore be a good default.',
                'de':'TwoPointsDE is excellent in many cases, including very high num_workers.',
                'portfolio':'PortfolioDiscreteOnePlusOne is excellent in discrete settings of mixed settings when high precision on parameters is not relevant. Its possibly a good choice for hyperparameter choice.',
                'oneplus':"OnePlusOne is a simple robust method for continuous parameters with num_workers less than 8",
                'cma':'CMA is excellent for control (e.g. neurocontrol) when the environment is not very noisy (num_workers ~50 ok) and when the budget is large (e.g. 1000 x the dimension).',
                'tbpsa':'TBPSA is excellent for problems corrupted by noise, in particular overparameterized (neural) ones; very high num_workers ok).',
                'pso':'PSO is excellent in terms of robustness, high num_workers ok.',
                'hammersley':'ScrHammersleySearchPlusMiddlePoint is excellent for super parallel cases (fully one-shot, i.e. num_workers = budget included) or for very multimodal cases (such as some of our MLDA problems); don’t use softmax with this optimizer.',
                'random':'RandomSearch is the classical random search baseline; don’t use softmax with this optimizer.'}


def nevergrad_cube(objective, n_trials,n_dim, with_count=False, method=None):

    instrument = ng.p.Array(shape=(n_dim,)).set_bounds(lower=0.0,upper=1.0)
    num_workers = 1

    if method.lower()=='ngopt':
        optimizer = ng.optimizers.NGOpt(parametrization=instrument, budget=n_trials, num_workers=num_workers)
    elif method.lower()=='ngopt4':
        optimizer = ng.optimizers.NGOpt4(parametrization=instrument, budget=n_trials,  num_workers=num_workers)
    elif method.lower() == 'ngopt8':
        optimizer = ng.optimizers.NGOpt8(parametrization=instrument, budget=n_trials,  num_workers=num_workers)
    elif method.lower() == 'de':
        optimizer = ng.optimizers.TwoPointsDE(parametrization=instrument, budget=n_trials, num_workers=num_workers)
    elif method.lower() == 'portfolio':
        optimizer = ng.optimizers.Portfolio(parametrization=instrument, budget=n_trials,  num_workers=num_workers)
    elif method.lower() == 'oneplus':
        optimizer = ng.optimizers.OnePlusOne(parametrization=instrument, budget=n_trials,  num_workers=num_workers)
    elif method.lower() == 'tbpsa':
        optimizer = ng.optimizers.TBPSA(parametrization=instrument, budget=n_trials, num_workers=num_workers)
    elif method.lower() == 'hammersley':
        optimizer = ng.optimizers.ScrHammersleySearchPlusMiddlePoint(parametrization=instrument, budget=n_trials,
                                        num_workers=num_workers)
    elif method.lower() == 'pso':
        optimizer = ng.optimizers.PSO(parametrization=instrument, budget=n_trials, num_workers=num_workers)
    elif method.lower() == 'cma':
        optimizer = ng.optimizers.CMA(parametrization=instrument, budget=n_trials, num_workers=num_workers)
    elif method.lower() == 'random':
        optimizer = ng.optimizers.RandomSearch(parametrization=instrument, budget=n_trials, num_workers=num_workers)
    else:
        raise Exception('Not recognizing '+str(method))
    global feval_count
    feval_count = 0

    def cube_objective(us):
        global feval_count
        feval_count += 1
        return objective(us)

    recommendation = optimizer.minimize( cube_objective )
    best_x = recommendation.value
    best_val = recommendation.loss

    if with_count:
        return best_val, best_x, feval_count
    else:
        return best_val, best_x


def nevergrad_ngopt_cube(objective, n_trials,n_dim, with_count=False):
    return nevergrad_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='ngopt')


def nevergrad_ngopt4_cube(objective, n_trials,n_dim, with_count=False):
    return nevergrad_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='ngopt4')


def nevergrad_ngopt8_cube(objective, n_trials,n_dim, with_count=False):
    return nevergrad_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='ngopt8')


def nevergrad_de_cube(objective, n_trials,n_dim, with_count=False):
    return nevergrad_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='de')


def nevergrad_portfolio_cube(objective, n_trials,n_dim, with_count=False):
    return nevergrad_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='portfolio')


def nevergrad_oneplus_cube(objective, n_trials,n_dim, with_count=False):
    return nevergrad_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='oneplus')


def nevergrad_cma_cube(objective, n_trials,n_dim, with_count=False):
    return nevergrad_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='cma')


def nevergrad_hammersley_cube(objective, n_trials,n_dim, with_count=False):
    return nevergrad_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='hammersley')


BAD = [ nevergrad_cma_cube ] # TODO: fix

NEVERGRAD_OPTIMIZERS = [ nevergrad_ngopt_cube, nevergrad_ngopt4_cube, nevergrad_ngopt8_cube,
                         nevergrad_de_cube, nevergrad_portfolio_cube, nevergrad_oneplus_cube,
                         nevergrad_hammersley_cube ]


if __name__=='__main__':
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in NEVERGRAD_OPTIMIZERS:
            print((optimizer.__name__,optimizer(objective, n_trials=250, n_dim=6, with_count=True)))



from scipy.optimize import shgo

# Define how SHGO does local search
MINIMIZER_KWARGS = {'slsqp': {'method': 'SLSQP',
                              'max_iter': 5},
                    'powell': {'method': 'Powell',
                               'max_iter': 5},
                    'nelder': {'method': 'Nelder-Mead',
                               'maxiter': 5}
                    }


def shgo_slsqp_sobol_cube(objective, n_trials, n_dim, with_count: bool = False):
    return shgo_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                     local_method='slsqp', sampling_method='sobol')


def shgo_slsqp_simplicial_cube(objective, n_trials, n_dim, with_count: bool = False):
    return shgo_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                     local_method='slsqp', sampling_method='simplicial')


def shgo_powell_sobol_cube(objective, n_trials, n_dim, with_count: bool = False):
    return shgo_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                     local_method='powell', sampling_method='sobol')


def shgo_powell_simplicial_cube(objective, n_trials, n_dim, with_count: bool = False):
    return shgo_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                     local_method='powell', sampling_method='simplicial')


def shgo_nelder_sobol_cube(objective, n_trials, n_dim, with_count: bool = False):
    return shgo_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                     local_method='nelder', sampling_method='sobol')


def shgo_nelder_simplicial_cube(objective, n_trials, n_dim, with_count: bool = False):
    return shgo_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                     local_method='nelder', sampling_method='simplicial')


SHGO_OPTIMIZERS_ALL = [shgo_slsqp_sobol_cube, shgo_slsqp_simplicial_cube, shgo_powell_sobol_cube,
                   shgo_powell_simplicial_cube, shgo_nelder_sobol_cube, shgo_nelder_simplicial_cube]

SHGO_OPTIMIZERS = [ o for o in SHGO_OPTIMIZERS_ALL if 'sobol' in o.__name__]


def shgo_cube(objective, n_trials, n_dim, with_count: bool = False, local_method=None, sampling_method='sobol'):
    """ Minimize a function on the cube using SHGO
    :param objective:    function on (0,1)^n_dim
    :param n_trials:
    :param n_dim:
    :param with_count:
    :return:
    """
    minimizer_kwargs = MINIMIZER_KWARGS[local_method]
    assert sampling_method in ['sobol', 'simplicial'], ' did not understand sampling method'
    bounds = [(0, 1)] * n_dim

    global feval_count
    feval_count = 0

    def _objective(x):
        global feval_count
        feval_count += 1
        return objective(list(x))

    # Try to induce roughly the right number of function evaluations. This can be improved!
    n_trials_reduced = int(n_trials/2+1)
    n_iters = int(1 + n_trials / 80)
    n = int(5 + n_trials / 40)
    result = shgo(_objective, bounds, n=n, iters=n_iters, options={'maxfev': n_trials_reduced,
                                                                   'minimize_every_iter': False,
                                                                   'maxfun': n_trials_reduced,
                                                                   'minimizer_kwargs': minimizer_kwargs},
                  sampling_method=sampling_method)
    return (result.fun, list(result.x), feval_count) if with_count else (result.fun, result.x)


if __name__ == '__main__':
    from humpday.objectives.classic import CLASSIC_OBJECTIVES
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in SHGO_OPTIMIZERS:
            print(optimizer(objective, n_trials=200, n_dim=5, with_count=True))



from scipy.optimize import minimize
from humpday.objectives.classic import CLASSIC_OBJECTIVES
global feval_count

# TODO: standardize with SHGO and remove duplicate code
from copy import deepcopy
MINIMIZER_KWARGS = {'slsqp': {'method': 'SLSQP'},
                    'powell': {'method': 'Powell'},
                    'nelder': {'method': 'Nelder-Mead'},
                    'dogleg': {'method': 'dogleg'},
                    'lbfgsb': {'method': 'L-BFGS-B'},
}


def scipy_cube(objective, n_trials, n_dim, with_count=False, method=None):
    bounds = [(0,1) ]*n_dim

    options = deepcopy(MINIMIZER_KWARGS[method])
    options.update({'maxfev': n_trials,'maxiter':n_trials})

    global feval_count
    feval_count = 0

    def _objective(x):
        global feval_count
        feval_count +=1
        return objective(list(x))

    result = minimize(_objective, x0=[0]*n_dim, method=options['method'],bounds=bounds, options=options)
    best_x = result.x.tolist()
    best_val = _objective(result.x)
    return (best_val, best_x,  feval_count) if with_count else (best_val, best_x)


def scipy_slsqp_cube(objective, n_trials, n_dim, with_count=False ):
    return scipy_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='slsqp')


def scipy_powell_cube(objective, n_trials, n_dim, with_count=False):
    return scipy_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='powell')


def scipy_nelder_cube(objective, n_trials, n_dim, with_count=False):
    return scipy_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='nelder')


def scipy_lbfgsb_cube(objective, n_trials, n_dim, with_count=False):
    return scipy_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='lbfgsb')


SCIPY_OPTIMIZERS = [ scipy_slsqp_cube, scipy_powell_cube, scipy_nelder_cube, scipy_lbfgsb_cube ]


if __name__ == '__main__':
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in SCIPY_OPTIMIZERS:
            print((optimizer.__name__,(optimizer(objective, n_trials=250, n_dim=6, with_count=True))))




import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="humpday",
    version="0.1.19",
    description="Taking the pain out of choosing a Python global optimizer",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/microprediction/humpday",
    author="microprediction",
    author_email="pcotton@intechinvestments.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["humpday", "humpday.objectives", "humpday.optimizers", "humpday.comparison", "humpday.visualization"],
    test_suite='pytest',
    tests_require=['pytest'],
    include_package_data=True,
    install_requires=["wheel", "pathlib", "numpy>=1.19.5", "importlib-metadata>=1.7.0", "getjson",
                      "scipy", "scikit-optimize", "poap", "pySOT", "funcy", "optuna", "deap", "ax-platform",
                      "platypus-opt",
                      "pymoo", "nevergrad", "landscapes", "bayesian-optimization", "cmake", "dlib"],
    entry_points={
        "console_scripts": [
            "humpday=humpday.__main__:main",
        ]
    },
)



from swarmlib.util.problem_base import ProblemBase
from swarmlib.pso.particle import Particle as PSOParticle
import numpy as np
from typing import Iterable, Tuple


# I like to give the little guys a run...but this isn't quite there.
# This rips out the guts of swarmlib so that it isn't bundled to visualization
# But it's still only good for 1,2,3 dim, and only really works for 2-dim.
# Here we convert 3-dim to 2-dim using space filling curve, which may not be the best idea.
# Hey, I tried.

class NoVisualizer():

    def __init__(self, **kwargs):
        self.__lower_boundary = kwargs.get('lower_boundary', 0.)
        self.__upper_boundary = kwargs.get('upper_boundary', 4.)
        self.__iteration_number = kwargs.get('iteration_number', 10)
        self.__intervals = self.__iteration_number + 2  # Two extra intervals for unanimated start and end pose
        self.__interval_ms = kwargs.get('interval', 1000)
        self.__continuous = kwargs.get('continuous', False)
        self._dark = kwargs.get('dark', False)

        self.__function = kwargs['function']

        self._marker_size = 0
        self._index = 0
        self._vel_color = '#CFCFCF'
        self._marker_color = '#0078D7' if self._dark else '#FF0000'
        self._marker_colors = np.empty(0)

        self._positions = []
        self._velocities = []
        self.__frame_interval = 50  # ms

    def add_data(self, **kwargs) -> None:
        positions: Iterable[Tuple[float, float]] = kwargs['positions']

        self._positions.append(np.transpose(positions))

        if len(self._positions) == 1:
            # Insert the first position twice to show it "unanimated" first.
            self._positions.append(np.transpose(positions))

        # Calculate at time t the velocity for step t-1
        self._velocities.append(self._positions[-1] - self._positions[-2])


class InvisiblePSOProblem(ProblemBase):
    def __init__(self, **kwargs):
        """
        Initialize a new particle swarm optimization problem.
        """
        super().__init__(**kwargs)
        self.__iteration_number = kwargs['iteration_number']
        self.__particles = [
            PSOParticle(**kwargs, bit_generator=self._random)
            for _ in range(kwargs['particles'])
        ]

        # The library stores particles in the visualizer .... groan
        positions = [particle.position for particle in self.__particles]
        self._visualizer = NoVisualizer(**kwargs)
        self._visualizer.add_data(positions=positions)

    def solve(self) -> PSOParticle:
        # And also update global_best_particle
        for _ in range(self.__iteration_number):

            # Update global best
            global_best_particle = min(self.__particles)

            for particle in self.__particles:
                particle.step(global_best_particle.position)

            # Add data for plot
            positions = [particle.position for particle in self.__particles]
            self._visualizer.add_data(positions=positions)

        return global_best_particle


def swarmlib_cube(objective, n_trials, n_dim, with_count=False, algo=None):
    """ Minimize a function on the cube using HyperOpt, and audit # of function calls
       :param objective:    function on (0,1)^n_dim
       :param n_trials:     Guideline for function evaluations
       :param n_dim:
       :param with_count:
       :return:
    """
    assert algo=='pso'
    assert n_dim==2,'yeah, sorry'

    global feval_count
    feval_count = 0

    def cube_objective(us):
        # PSO only handles 2-dim
        assert all( [ 0<=ui<=1 for ui in us]),' expecting value on square '

        global feval_count
        feval_count +=1
        return objective(us)

    iteration_number = 5 if n_trials < 50 else 10
    particles = max( int( n_trials / iteration_number), 1)
    problem = InvisiblePSOProblem(function=cube_objective, particles=particles, iteration_number=iteration_number,
                         lower_boundary=0., upper_boundary=1.0)
    best_particle = problem.solve()
    best_x = best_particle.position.tolist()
    best_val = best_particle.value
    return (best_val, best_x,  feval_count) if with_count else (best_val, best_x)


def swarmlib_pso_cube(objective, n_trials, n_dim, with_count=False):
    return swarmlib_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, algo='pso')


SWARMLIB_OPTIZERS = [] # Not ready for the A-league yet.


if __name__ == '__main__':
    from humpday.objectives.classic import CLASSIC_OBJECTIVES

    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for n_dim in range(2,4):
            print('n_dim='+str(n_dim))
            for optimizer in SWARMLIB_OPTIZERS:
                print(optimizer(objective, n_trials=100, n_dim=n_dim, with_count=True))
